from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Protocol

import rich

from pdm_shear.parser import ImportStatement


class FixAction(Protocol):
    def __call__(self, data: dict[str, Any]) -> None:
        """Fix the data in place."""
        ...


@dataclass(frozen=True)
class UnusedDependency:
    """Remove a dependency from the project."""

    dependency: str

    def __call__(self, data: dict[str, Any]) -> None:
        if (
            "dependencies" in data.get("project", {})
            and self.dependency in data["project"]["dependencies"]
        ):
            data["project"]["dependencies"].remove(self.dependency)
            rich.print(f"Removed dependency: [green bold]{self.dependency}[/]")

    def __rich__(self) -> str:
        return f"Found unused dependency: [green bold]{self.dependency}[/]"


@dataclass(frozen=True)
class UnusedOptionalDependency:
    """Remove an optional dependency from the project."""

    dependency: str
    group: str

    def __call__(self, data: dict[str, Any]) -> None:
        if (
            "optional-dependencies" in data.get("project", {})
            and self.group in data["project"]["optional-dependencies"]
        ):
            group = data["project"]["optional-dependencies"][self.group]
            if self.dependency in group:
                group.remove(self.dependency)
                rich.print(
                    f"Removed optional dependency of group [yellow]{self.group}[/]: [green bold]{self.dependency}[/]"
                )

    def __rich__(self) -> str:
        return f"Found unused optional dependency of group [yellow]{self.group}[/]: [green bold]{self.dependency}[/]"


@dataclass(frozen=True)
class MissingDependency:
    """Missing dependency."""

    module_name: str
    statements: list[ImportStatement]

    def __rich__(self) -> str:
        text = f"Missing dependency: [green bold]{self.module_name}[/] used by the following files:"
        for stm in self.statements:
            text += f"\n{stm.__rich__()}"
        return text

    def __call__(self, data: dict[str, Any]) -> None:
        rich.print(self)



import pytest


@pytest.fixture
def mock_data(monkeypatch):
    data = {
        "requests": ["requests"],
        "flask": ["flask"],
        "python-multipart": ["python_multipart", "multipart"],
        "django": ["django"],
        "rich": ["rich"],
    }
    monkeypatch.setattr("pdm_shear.main._local_cache", data)



from __future__ import annotations

import ast
from dataclasses import dataclass
from fnmatch import fnmatchcase
from pathlib import Path
from typing import Iterable


@dataclass(frozen=True)
class ImportStatement:
    module: str
    file_path: Path
    line_number: int
    col_offset: int
    end_col_offset: int

    def _comp_key(self) -> tuple[str, str, int, int]:
        """Comparison key for sorting."""
        return (
            self.module,
            self.file_path.as_posix(),
            self.line_number,
            self.col_offset,
        )

    def __lt__(self, other: ImportStatement) -> bool:
        """Less than comparison for sorting."""
        return self._comp_key() < other._comp_key()

    def __rich__(self) -> str:
        """Rich representation for display."""
        with open(self.file_path) as file:
            line = file.readlines()[self.line_number - 1].rstrip()
        rendered = f"{line[: self.col_offset]}[red]{line[self.col_offset : self.end_col_offset]}[/]{line[self.end_col_offset :]}"
        return f"{self.file_path}:{self.line_number}\n{rendered}"


def _parse_imports_single_file(file_path: Path) -> Iterable[ImportStatement]:
    """Parse the given Python source file to detect imports."""

    with open(file_path, "r") as file:
        node = ast.parse(file.read(), filename=file_path)

    for n in ast.walk(node):
        if isinstance(n, ast.Import):
            for alias in n.names:
                if not alias.name:
                    continue
                yield ImportStatement(
                    module=alias.name.split(".")[0],
                    file_path=file_path,
                    line_number=alias.lineno,
                    col_offset=alias.col_offset,
                    end_col_offset=alias.end_col_offset,
                )
        elif (
            isinstance(n, ast.ImportFrom) and n.module
        ):  # relative import may not have module name
            yield ImportStatement(
                module=n.module.split(".")[0],
                file_path=file_path,
                line_number=n.lineno,
                col_offset=n.col_offset,
                end_col_offset=n.end_col_offset,
            )


DEFAULT_EXCLUDE = [".venv/*"]


def get_project_imports(
    project_path: Path,
    include: list[str] | None = None,
    exclude: list[str] | None = None,
) -> Iterable[ImportStatement]:
    """Get all imports from the project directory."""
    if not project_path.is_dir():
        raise ValueError(f"Path {project_path} is not a directory.")
    for path in project_path.rglob("*.py"):
        rel_path = path.relative_to(project_path).as_posix()
        if any(fnmatchcase(rel_path, pattern) for pattern in DEFAULT_EXCLUDE):
            continue
        if include is not None and not any(
            fnmatchcase(rel_path, pattern) for pattern in include
        ):
            continue
        if exclude is not None and any(
            fnmatchcase(rel_path, pattern) for pattern in exclude
        ):
            continue
        yield from _parse_imports_single_file(path)



from __future__ import annotations

import csv
import json
import subprocess
import zipfile
from itertools import groupby
from typing import TYPE_CHECKING, Iterable

import rich

from pdm_shear.actions import (
    FixAction,
    MissingDependency,
    UnusedDependency,
    UnusedOptionalDependency,
)
from pdm_shear.parser import get_project_imports

if TYPE_CHECKING:
    from pdm.project import Project


def filter_site_packages(project: Project, module_names: list[str]) -> list[str]:
    """Return a list of module names that are installed in site-packages."""
    script = f"""\
import json
import importlib.util
module_names = {module_names!r}
site_packages = []
for module_name in module_names:
    try:
        spec = importlib.util.find_spec(module_name)
    except ModuleNotFoundError:
        site_packages.append(module_name)
        continue
    if not spec or spec.origin and "site-packages" in spec.origin:
        site_packages.append(module_name)
print(json.dumps(site_packages))
"""
    try:
        result = subprocess.run(
            [str(project.environment.interpreter.path), "-c", script],
            capture_output=True,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"Failed to run Python script: {e.stderr}") from e
    else:
        return json.loads(result.stdout)


_local_cache: dict[str, list[str]] = {}


def get_top_level_names(project: Project, dependency: str) -> list[str]:
    """Get the top-level names of the distribution."""
    from pdm.exceptions import ProjectError
    from pdm.models.caches import JSONFileCache
    from pdm.models.candidates import Candidate
    from pdm.models.requirements import parse_requirement

    req = parse_requirement(dependency)
    working_set = project.environment.get_working_set()
    dist_name = req.key
    if not dist_name:
        raise ProjectError(f"Dependency name must be given: {dependency!r}")
    # 0. Check local cache
    if dist_name in _local_cache:
        return _local_cache[dist_name]
    # 1. Get from installed package
    if dist_name in working_set:
        dist = working_set[dist_name]
        if top_level := dist.read_text("top_level.txt"):
            result = top_level.splitlines()
        else:
            result = _collect_top_level_names(
                path.as_posix() for path in dist.files or []
            )
        _local_cache[dist_name] = result
        return result
    # 2. Get from cache file
    cache = JSONFileCache(project.cache_dir / "top_level_names.json")
    if project.core.state.enable_cache and dist_name in cache:
        cached = cache.get(dist_name)
        _local_cache[dist_name] = cached
        return cached
    # 3. Get from PyPI wheel
    with project.environment.get_finder() as finder:
        best_match = finder.find_best_match(dependency)
        if not best_match.best:
            raise ProjectError(f"Cannot find distribution for {dependency!r}")
        candidate = Candidate.from_installation_candidate(best_match.best, req)
        prepared = candidate.prepare(project.environment)
        wheel = prepared.build()
        with zipfile.ZipFile(wheel, "r") as zf:
            record_file = next(
                f for f in zf.namelist() if f.endswith(".dist-info/RECORD")
            )

            record_lines = zf.read(record_file).decode("utf-8").splitlines()
            top_level_names = _collect_top_level_names(
                name for name, *_ in csv.reader(record_lines)
            )
            cache.set(dist_name, top_level_names)
            _local_cache[dist_name] = top_level_names
            return top_level_names


def _collect_top_level_names(paths: Iterable[str]) -> list[str]:
    top_level_names: set[str] = set()
    for path in paths:
        if path.startswith("."):
            continue
        if not path.endswith((".py", ".pyi", ".pyc", ".pyd", ".pyo", ".so")):
            continue
        top_name = path.split("/")[0].split(".")[0]
        top_level_names.add(top_name)
    return list(top_level_names)


def fix_or_show_unused_dependencies(
    project: Project, fix: bool = False, ignore_missing_deps: bool = False
) -> bool:
    from pdm.utils import normalize_name

    data = project.pyproject.read()
    tool_config = data.get("tool", {}).get("pdm", {}).get("shear", {})
    imports = sorted(
        get_project_imports(
            project.root, tool_config.get("include"), tool_config.get("exclude")
        )
    )
    grouped = {k: list(v) for k, v in groupby(imports, key=lambda x: x.module)}
    used_names = filter_site_packages(project, list(grouped.keys()))
    used_grouped = {k: v for k, v in grouped.items() if k in used_names}

    actions: list[FixAction] = []
    for dep in project.pyproject.metadata.get("dependencies", []):
        top_names = get_top_level_names(project, dep)
        if set(top_names).isdisjoint(used_names):
            actions.append(UnusedDependency(dependency=dep))
        else:
            for name in top_names:
                used_grouped.pop(name, None)

    for group, deps in project.pyproject.metadata.get(
        "optional-dependencies", {}
    ).items():
        for dep in deps:
            top_names = get_top_level_names(project, dep)
            if set(top_names).isdisjoint(used_names):
                actions.append(UnusedOptionalDependency(dependency=dep, group=group))
            else:
                for name in top_names:
                    used_grouped.pop(name, None)
    if not ignore_missing_deps and not tool_config.get("ignore_missing_deps", False):
        for name, statements in used_grouped.items():
            if project.name and normalize_name(project.name) == name:
                continue
            actions.append(
                MissingDependency(module_name=name, statements=sorted(statements))
            )

    if fix:
        for action in actions:
            action(data)
        project.pyproject.set_data(data)
        project.pyproject.write()
    else:
        for action in actions:
            rich.print(action)
    if any(isinstance(action, MissingDependency) for action in actions):
        rich.print(
            "[yellow]Add them to the dependencies or ignore the warning with `--ignore-missing-deps`[/]"
        )
    return len(actions) == 0



"""
pdm_shear

:Please add description here:
"""

from __future__ import annotations

from argparse import ArgumentParser, Namespace

from pdm.cli.commands.base import BaseCommand
from pdm.core import Core
from pdm.project import Project


class ShearCommand(BaseCommand):
    """Detect and remove unused dependencies for Python projects"""

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "--fix",
            action="store_true",
            help="Fix the pyproject.toml to remove the unused dependencies",
        )
        parser.add_argument(
            "--ignore-missing-deps",
            action="store_true",
            help="Ignore missing dependencies",
        )

    def handle(self, project: Project, options: Namespace) -> None:
        from pdm_shear.main import fix_or_show_unused_dependencies

        result = fix_or_show_unused_dependencies(
            project, options.fix, options.ignore_missing_deps
        )
        raise SystemExit(int(not result))


def plugin(core: Core) -> None:
    """Register pdm plugin to pdm-core."""
    core.register_command(ShearCommand, "shear")



import textwrap

pytest_plugins = ["pdm.pytest"]


def test_unused_deps(pdm, mock_data, project):
    project.pyproject.metadata["dependencies"] = ["requests", "flask"]
    project.pyproject.metadata["optional-dependencies"] = {
        "web": ["django"],
    }
    project.pyproject.write()
    project.root.joinpath("test.py").write_text(
        textwrap.dedent(
            """\
        import requests as req
        from django.conf import settings
        import rich
        """
        )
    )
    result = pdm(["shear"], obj=project)
    assert result.exit_code == 1
    assert "Found unused dependency: flask" in result.stdout
    assert "Found unused dependency: requests" not in result.stdout
    assert "Found unused optional dependency of group web: django" not in result.stdout

    result = pdm(["shear", "--fix"], obj=project)
    assert result.exit_code == 1
    assert "Removed dependency: flask" in result.stdout
    assert project.pyproject.metadata["dependencies"] == ["requests"]


def test_missing_deps(pdm, mock_data, project):
    project.pyproject.metadata["dependencies"] = ["requests"]
    project.pyproject.write()
    project.root.joinpath("test.py").write_text(
        textwrap.dedent(
            """\
        import requests
        from django.conf import settings
        import rich
        """
        )
    )
    result = pdm(["shear"], obj=project)
    assert result.exit_code == 1
    assert "Missing dependency: django" in result.stdout
    assert "Missing dependency: rich" in result.stdout

    result = pdm(["shear", "--ignore-missing-deps"], obj=project)
    assert result.exit_code == 0

    project.pyproject._data.setdefault("tool", {}).setdefault("pdm", {})["shear"] = {
        "ignore_missing_deps": True
    }
    project.pyproject.write()
    result = pdm(["shear"], obj=project)
    assert result.exit_code == 0
